import os
import sys

sys.path.append(os.path.abspath(''))

from brain_brew.commands.run_recipe.top_level_builder import TopLevelBuilder

build: str = TopLevelBuilder.build_yamale()
filepath = "brain_brew/schemas/recipe.yaml"

with open(filepath, 'w') as fp:
    fp.write(build)
    fp.close()



from dataclasses import dataclass, field
from typing import Dict, List, Union, Optional

from brain_brew.build_tasks.csvs.shared_base_csvs import SharedBaseCsvs
from brain_brew.commands.run_recipe.build_task import BuildPartTask
from brain_brew.configuration.part_holder import PartHolder
from brain_brew.representation.yaml.notes import Note, Notes
from brain_brew.transformers.file_mapping import FileMapping
from brain_brew.transformers.note_model_mapping import NoteModelMapping
from brain_brew.utils import split_tags


@dataclass
class NotesFromCsvs(SharedBaseCsvs, BuildPartTask):
    @classmethod
    def task_name(cls) -> str:
        return r'notes_from_csvs'

    @classmethod
    def task_regex(cls) -> str:
        return r'notes_from_csv[s]?'

    @classmethod
    def yamale_schema(cls) -> str:
        return f'''\
            part_id: str()
            save_to_file: str(required=False)
            note_model_mappings: list(include('{NoteModelMapping.task_name()}'))
            file_mappings: list(include('{FileMapping.task_name()}'))
        '''

    @classmethod
    def yamale_dependencies(cls) -> set:
        return {NoteModelMapping, FileMapping}

    @dataclass
    class Representation(SharedBaseCsvs.Representation):
        part_id: str
        save_to_file: Optional[str] = field(default=None)

        def encode(self):
            return {
                "part_id": self.part_id,
                "save_to_file": self.save_to_file,
                "file_mappings": [fm.encode() for fm in self.file_mappings],
                "note_model_mappings": [nmm.encode() for nmm in self.note_model_mappings]
            }

    @classmethod
    def from_repr(cls, data: Union[Representation, dict]):
        rep: cls.Representation = data if isinstance(data, cls.Representation) else cls.Representation.from_dict(data)
        return cls(
            rep=rep,
            part_id=rep.part_id,
            save_to_file=rep.save_to_file,
            file_mappings=rep.get_file_mappings(),
            note_model_mappings={k: v for nm in rep.note_model_mappings for k, v in cls.map_nmm(nm).items()}
        )

    rep: Representation
    part_id: str
    save_to_file: Optional[str]

    def execute(self):
        self.verify_contents()

        csv_data_by_guid: Dict[str, dict] = {}
        for csv_map in self.file_mappings:
            csv_map.compile_data()
            csv_data_by_guid = {**csv_data_by_guid, **csv_map.compiled_data}
            csv_map.write_file_on_close()
        csv_rows: List[dict] = list(csv_data_by_guid.values())

        notes_part: List[Note] = [self.csv_row_to_note(row, self.note_model_mappings) for row in csv_rows]

        notes = Notes.from_list_of_notes(notes_part)
        PartHolder.override_or_create(self.part_id, self.save_to_file, notes)

    @staticmethod
    def csv_row_to_note(row: dict, note_model_mappings: Dict[str, NoteModelMapping]) -> Note:
        note_model_name = row["note_model"]  # TODO: Use object
        row_nm: NoteModelMapping = note_model_mappings[note_model_name]

        filtered_fields = row_nm.csv_row_map_to_note_fields(row)

        guid = filtered_fields.pop("guid")
        tags = split_tags(filtered_fields.pop("tags"))
        flags = filtered_fields.pop("flags") if "flags" in filtered_fields else 0

        fields = row_nm.field_values_in_note_model_order(note_model_name, filtered_fields)

        return Note(guid=guid, tags=tags, note_model=note_model_name, fields=fields, flags=flags)



import glob
import logging
from typing import List

from brain_brew.representation.generic.source_file import SourceFile
from brain_brew.representation.json.json_file import JsonFile
from brain_brew.representation.json.wrappers_for_crowd_anki import CrowdAnkiJsonWrapper
from brain_brew.representation.yaml.note_model import NoteModel
from brain_brew.utils import create_path_if_not_exists


class CrowdAnkiExport(SourceFile):
    folder_location: str
    json_file_location: str
    # import_config: CrowdAnkiImportConfig  # TODO: Make this
    json_data: CrowdAnkiJsonWrapper
    note_models: List[NoteModel]

    media_loc: str

    def __init__(self, folder_location):
        self.folder_location = folder_location
        if self.folder_location[-1] != "/":
            self.folder_location = self.folder_location + "/"

        create_path_if_not_exists(self.folder_location)

        self.json_file_location = self.find_json_file_in_folder()
        self._read_json_file()

        self.media_loc = self.folder_location + "media/"

        if not self.is_dir(self.media_loc):
            create_path_if_not_exists(self.media_loc)
            return

    @classmethod
    def from_file_loc(cls, file_loc) -> 'CrowdAnkiExport':
        return cls(file_loc)

    def find_json_file_in_folder(self):
        files = glob.glob(f"{glob.escape(self.folder_location)}*.json")

        if len(files) == 1:
            return files[0]
        elif not files:
            file_loc = self.folder_location + "deck.json"
            logging.warning(f"Creating missing json file '{file_loc}'")
            return file_loc
        else:
            logging.error(f"Multiple json files found in '{self.folder_location}': {files}")
            raise FileExistsError()

    def write_to_files(self, json_data):  # import_config_data
        JsonFile.write_file(self.json_file_location, json_data)

    def _read_json_file(self):
        if SourceFile.is_file(self.json_file_location):
            self.json_data = CrowdAnkiJsonWrapper(JsonFile.read_file(self.json_file_location))
            self.note_models = list(map(NoteModel.from_crowdanki, self.json_data.note_models))
        else:
            self.write_to_files({})
            self.json_data = CrowdAnkiJsonWrapper({})



import copy
from pathlib import Path


class SourceFile(object):
    @classmethod
    def from_file_loc(cls, file_loc) -> 'SourceFile':
        pass

    @classmethod
    def is_file(cls, filename: str):
        return Path(filename).is_file()

    @classmethod
    def is_dir(cls, folder_name: str):
        return Path(folder_name).is_dir()

    @classmethod
    def get_deep_copy(cls, data):
        return copy.deepcopy(data)

    @classmethod
    def create_or_get(cls, location):
        from brain_brew.configuration.file_manager import FileManager
        _file_manager = FileManager.get_instance()
        formatted_location = cls.formatted_file_location(location)
        file = _file_manager.file_if_exists(formatted_location)

        if file is not None:
            return file

        # if not cls.is_file(location) and not cls.is_dir(location):
        #     raise FileNotFoundError(f"No file or folder '{location}' exists")

        file = cls.from_file_loc(location)
        _file_manager.register_file(formatted_location, file)
        return file

    @classmethod
    def formatted_file_location(cls, location):
        return location



import logging
from dataclasses import dataclass, field
from typing import Union, Optional, List

from brain_brew.build_tasks.crowd_anki.shared_base_notes import SharedBaseNotes
from brain_brew.commands.run_recipe.build_task import BuildPartTask
from brain_brew.configuration.part_holder import PartHolder
from brain_brew.configuration.representation_base import RepresentationBase
from brain_brew.representation.json.crowd_anki_export import CrowdAnkiExport
from brain_brew.representation.json.wrappers_for_crowd_anki import CrowdAnkiJsonWrapper, CrowdAnkiNoteWrapper
from brain_brew.representation.yaml.notes import Notes, Note


@dataclass
class NotesFromCrowdAnki(SharedBaseNotes, BuildPartTask):
    @classmethod
    def task_name(cls) -> str:
        return r'notes_from_crowd_anki'

    @classmethod
    def yamale_schema(cls) -> str:
        return f'''\
            part_id: str()
            source: str()
            sort_order: list(str(), required=False)
            save_to_file: str(required=False)
            reverse_sort: str(required=False)
        '''

    @dataclass
    class Representation(RepresentationBase):
        part_id: str
        source: str
        sort_order: Optional[List[str]] = field(default_factory=lambda: None)
        reverse_sort: Optional[bool] = field(default_factory=lambda: None)
        save_to_file: Optional[str] = field(default=None)

    @classmethod
    def from_repr(cls, data: Union[Representation, dict]):
        rep: cls.Representation = data if isinstance(data, cls.Representation) else cls.Representation.from_dict(data)
        return cls(
            rep=rep,
            ca_export=CrowdAnkiExport.create_or_get(rep.source),
            part_id=rep.part_id,
            sort_order=SharedBaseNotes._get_sort_order(rep.sort_order),
            reverse_sort=SharedBaseNotes._get_reverse_sort(rep.reverse_sort),
            save_to_file=rep.save_to_file
        )

    rep: Representation
    part_id: str
    ca_export: CrowdAnkiExport
    sort_order: Optional[List[str]]
    reverse_sort: Optional[bool]
    save_to_file: Optional[str]

    def execute(self) -> PartHolder[Notes]:
        ca_wrapper: CrowdAnkiJsonWrapper = self.ca_export.json_data
        if ca_wrapper.children:
            logging.warning("Child Decks / Sub-decks are not currently supported.")

        ca_models = self.ca_export.note_models
        ca_notes = ca_wrapper.notes

        nm_id_to_name: dict = {model.id: model.name for model in ca_models}
        note_list = [self.ca_note_to_note(note, nm_id_to_name) for note in ca_notes]

        notes = Notes.from_list_of_notes(note_list)  # TODO: pass in sort method
        return PartHolder.override_or_create(self.part_id, self.save_to_file, notes)

    @staticmethod
    def ca_note_to_note(note: dict, nm_id_to_name: dict) -> Note:
        wrapper = CrowdAnkiNoteWrapper(note)

        return Note(
            note_model=nm_id_to_name[wrapper.note_model],
            tags=wrapper.tags,
            guid=wrapper.guid,
            fields=wrapper.fields,
            flags=wrapper.flags
        )



from dataclasses import dataclass
from brain_brew.interfaces.command import Command
from brain_brew.commands.run_recipe.top_level_builder import TopLevelBuilder
from brain_brew.configuration.yaml_verifier import YamlVerifier


@dataclass
class RunRecipe(Command):
    recipe_file_name: str
    verify_only: bool

    def execute(self):
        # Parse Build Config File
        YamlVerifier()
        recipe = TopLevelBuilder.parse_and_read(self.recipe_file_name, self.verify_only)

        if not self.verify_only:
            recipe.execute()



from dataclasses import dataclass
from typing import Union, List, Set

from brain_brew.configuration.part_holder import PartHolder
from brain_brew.configuration.representation_base import RepresentationBase
from brain_brew.interfaces.yamale_verifyable import YamlRepr
from brain_brew.representation.generic.media_file import MediaFile
from brain_brew.representation.yaml.media_group import MediaGroup
from brain_brew.transformers.save_media_group_to_location import save_media_groups_to_location


@dataclass
class MediaGroupToCrowdAnki(YamlRepr):
    @classmethod
    def task_name(cls) -> str:
        return r'media_group_to_crowd_anki'

    @classmethod
    def yamale_schema(cls) -> str:
        return f'''\
            parts: list(str())
        '''

    @dataclass
    class Representation(RepresentationBase):
        parts: List[str]

    @classmethod
    def from_repr(cls, data: Union[Representation, dict]):
        rep: cls.Representation = data if isinstance(data, cls.Representation) else cls.Representation.from_dict(data)
        return cls(
            rep=rep,
            parts=list(holder.part for holder in map(PartHolder.from_file_manager, rep.parts))
        )

    rep: Representation
    parts: List[MediaGroup]

    def execute(self, ca_media_folder: str) -> Set[MediaFile]:
        return save_media_groups_to_location(self.parts, ca_media_folder, True, False)



import logging
from dataclasses import dataclass, field
from typing import Optional, Union, List

from brain_brew.commands.run_recipe.build_task import BuildPartTask
from brain_brew.configuration.part_holder import PartHolder
from brain_brew.configuration.representation_base import RepresentationBase
from brain_brew.representation.json.crowd_anki_export import CrowdAnkiExport
from brain_brew.representation.json.wrappers_for_crowd_anki import CrowdAnkiJsonWrapper
from brain_brew.representation.yaml.note_model import NoteModel


@dataclass
class NoteModelsAllFromCrowdAnki(BuildPartTask):
    @classmethod
    def task_name(cls) -> str:
        return r'note_models_all_from_crowd_anki'

    @classmethod
    def yamale_schema(cls) -> str:
        return f'''\
            source: str()
        '''

    @dataclass
    class Representation(RepresentationBase):
        source: str

    @classmethod
    def from_repr(cls, data: Union[Representation, dict]):
        rep: cls.Representation = data if isinstance(data, cls.Representation) else cls.Representation.from_dict(data)
        return cls(
            rep=rep,
            ca_export=CrowdAnkiExport.create_or_get(rep.source)
        )

    rep: Representation
    ca_export: CrowdAnkiExport

    def execute(self) -> List[PartHolder[NoteModel]]:
        ca_wrapper: CrowdAnkiJsonWrapper = self.ca_export.json_data

        note_models_dict = {model.get('name'): model for model in ca_wrapper.note_models}

        parts = []
        for name, model in note_models_dict.items():
            parts.append(PartHolder.override_or_create(name, None, NoteModel.from_crowdanki(model)))

        logging.info(f"Found {len(parts)} note model{'s' if len(parts) > 1 else ''} in CrowdAnki Export: '"
                     + "', '".join(note_models_dict.keys()) + "'")

        return parts



import codecs
from dataclasses import dataclass

from brain_brew.representation.generic.source_file import SourceFile

_encoding = "utf-8"

@dataclass
class HTMLFile(SourceFile):
    file_location: str
    _data: str

    def __init__(self, file):
        self.file_location = file
        self.read_file()

    @classmethod
    def from_file_loc(cls, file_loc) -> 'HTMLFile':
        return cls(file_loc)

    def read_file(self):
        r = codecs.open(self.file_location, 'r', encoding=_encoding)
        self._data = r.read()

    def get_data(self, deep_copy=False) -> str:
        return self.get_deep_copy(self._data) if deep_copy else self._data

    @staticmethod
    def write_file(file_location, data):
        with open(file_location, "w+", encoding=_encoding) as file:
            file.write(data)

    @staticmethod
    def to_filename_html(filename: str) -> str:
        return filename + ".html" if not filename.endswith(".html") else filename

    @classmethod
    def formatted_file_location(cls, location):
        return cls.to_filename_html(location)



# class TestConstructor:
#     def test_runs(self):
#         with patch.object(CsvsGenerate, "__init__", return_value=None) as mock_csv_tr, \
#                 patch.object(DeckPartHolder, "from_part_pool", return_value=Mock()), \
#                 patch.object(CsvFile, "create_or_get", return_value=Mock()):
#
#             data = YamlObject.read_to_dict(TestFiles.BuildConfig.ONE_OF_EACH_TYPE)
#             builder = TopLevelRecipeBuilder.from_list(data)
#             builder.execute()
#
#             assert len(builder.tasks) == 1
#             assert mock_csv_tr.call_count == 1



import pytest

from brain_brew.representation.generic.csv_file import CsvFile
from tests.test_file_manager import get_new_file_manager
from tests.test_files import TestFiles

get_new_file_manager()


@pytest.fixture()
def csv_test1():
    csv = CsvFile(TestFiles.CsvFiles.TEST1)
    csv.read_file()
    return csv


@pytest.fixture()
def tsv_test1():
    tsv = CsvFile(TestFiles.TsvFiles.TEST1, delimiter='\t')
    tsv.read_file()
    return tsv


@pytest.fixture()
def csv_test1_split1():
    csv = CsvFile(TestFiles.CsvFiles.TEST1_SPLIT1)
    csv.read_file()
    return csv


@pytest.fixture()
def csv_test1_split2():
    csv = CsvFile(TestFiles.CsvFiles.TEST1_SPLIT2)
    csv.read_file()
    return csv


@pytest.fixture()
def csv_test2():
    csv = CsvFile(TestFiles.CsvFiles.TEST2)
    csv.read_file()
    return csv


@pytest.fixture()
def csv_test3():
    csv = CsvFile(TestFiles.CsvFiles.TEST3)
    csv.read_file()
    return csv


@pytest.fixture()
def csv_test2_missing_guids():
    csv = CsvFile(TestFiles.CsvFiles.TEST2_MISSING_GUIDS)
    csv.read_file()
    return csv


@pytest.fixture()
def temp_csv_test1(tmpdir, csv_test1) -> CsvFile:
    file = tmpdir.mkdir("json").join("file.csv")
    file.write("blank")

    csv = CsvFile.create_or_get(file.strpath)
    csv.read_file()
    return csv


class TestConstructor:
    def test_runs(self, csv_test1):
        assert isinstance(csv_test1, CsvFile)
        assert csv_test1.file_location == TestFiles.CsvFiles.TEST1
        assert "guid" in csv_test1.column_headers


def test_to_filename_csv():
    assert "read-this-file.csv" == CsvFile.to_filename_csv("read-this-file")
    assert "read-this-file.csv" == CsvFile.to_filename_csv("read-this-file.csv")
    assert "read-this-file.tsv" == CsvFile.to_filename_csv("read-this-file.tsv")


class TestWriteFile:
    def test_runs(self, temp_csv_test1: CsvFile, csv_test1: CsvFile):
        temp_csv_test1.set_data(csv_test1.get_data())
        temp_csv_test1.write_file()
        temp_csv_test1.read_file()

        assert temp_csv_test1.get_data() == csv_test1.get_data()

    def test_tsv_same_data(self, temp_csv_test1: CsvFile, tsv_test1: CsvFile):
        temp_csv_test1.set_data(tsv_test1.get_data())
        temp_csv_test1.write_file()
        temp_csv_test1.read_file()

        assert temp_csv_test1.get_data() == tsv_test1.get_data()


class TestSortData:
    @pytest.mark.parametrize("columns, reverse, result_column, expected_results", [
        (["guid"], False, "guid", [(0, "AAAA"), (1, "BBBB"), (2, "CCCC"), (14, "OOOO")]),
        (["guid"], True, "guid", [(14, "AAAA"), (13, "BBBB"), (12, "CCCC"), (0, "OOOO")]),
        (["english"], False, "english", [(0, "banana"), (1, "bird"), (2, "cat"), (14, "you")]),
        (["english"], True, "english", [(14, "banana"), (13, "bird"), (12, "cat"), (0, "you")]),
        (["tags"], False, "tags", [(0, "besttag"), (1, "funny"), (2, "tag2 tag3"), (13, ""), (14, "")]),
        (["esperanto", "english"], False, "esperanto", [(0, "banano"), (1, "birdo"), (6, "vi"), (14, "")]),
        (["esperanto", "guid"], False, "guid", [(7, "BBBB"), (14, "LLLL")]),
    ])
    def test_sort(self, csv_test1: CsvFile, columns, reverse, result_column, expected_results):
        csv_test1.sort_data(columns, reverse, case_insensitive_sort=True)

        sorted_data = csv_test1.get_data()

        for result in expected_results:
            assert sorted_data[result[0]][result_column] == result[1]

    def test_insensitive(self):
        pass



from dataclasses import dataclass, field
from typing import List, Union, Optional, Dict

from brain_brew.commands.run_recipe.build_task import TopLevelBuildTask
from brain_brew.configuration.part_holder import PartHolder
from brain_brew.configuration.representation_base import RepresentationBase
from brain_brew.representation.yaml.note_model import NoteModel
from brain_brew.transformers.save_note_model_to_location import save_note_model_to_location


@dataclass
class SaveNoteModelsToFolder(TopLevelBuildTask):
    @classmethod
    def task_name(cls) -> str:
        return r'save_note_models_to_folder'

    @classmethod
    def task_regex(cls) -> str:
        return r"save_note_models[s]?_to_folder"

    @classmethod
    def yamale_schema(cls) -> str:
        return f'''\
            parts: list(str())
            folder: str()
            clear_existing: bool(required=False)
        '''

    @dataclass
    class Representation(RepresentationBase):
        parts: List[str]
        folder: str
        clear_existing: Optional[bool] = field(default=None)

    @classmethod
    def from_repr(cls, data: Union[Representation, dict]):
        rep: cls.Representation = data if isinstance(data, cls.Representation) else cls.Representation.from_dict(data)
        return cls(
            rep=rep,
            parts=list(holder.part for holder in map(PartHolder.from_file_manager, rep.parts)),
            folder=rep.folder,
            clear_existing=rep.clear_existing or False,
        )

    rep: Representation
    parts: List[NoteModel]
    folder: str
    clear_existing: bool

    def execute(self) -> Dict[str, str]:
        model_yaml_files: Dict[str, str] = {}
        for model in self.parts:
            model_yaml_files.setdefault(
                model.name,
                save_note_model_to_location(model, self.folder, self.clear_existing)
            )
        return model_yaml_files



from dataclasses import dataclass, field
from typing import Union, Optional, List, Set

from brain_brew.build_tasks.crowd_anki.headers_to_crowd_anki import HeadersToCrowdAnki
from brain_brew.build_tasks.crowd_anki.media_group_to_crowd_anki import MediaGroupToCrowdAnki
from brain_brew.build_tasks.crowd_anki.note_models_to_crowd_anki import NoteModelsToCrowdAnki
from brain_brew.build_tasks.crowd_anki.notes_to_crowd_anki import NotesToCrowdAnki
from brain_brew.commands.run_recipe.build_task import TopLevelBuildTask
from brain_brew.configuration.representation_base import RepresentationBase
from brain_brew.representation.generic.media_file import MediaFile
from brain_brew.representation.json.crowd_anki_export import CrowdAnkiExport
from brain_brew.representation.json.wrappers_for_crowd_anki import CrowdAnkiJsonWrapper


@dataclass
class CrowdAnkiGenerate(TopLevelBuildTask):
    @classmethod
    def task_name(cls) -> str:
        return r'generate_crowd_anki'

    @classmethod
    def yamale_schema(cls) -> str:
        return f'''\
            folder: str()
            headers: str()
            notes: include('{NotesToCrowdAnki.task_name()}')
            note_models: include('{NoteModelsToCrowdAnki.task_name()}')
            media: include('{MediaGroupToCrowdAnki.task_name()}', required=False)
        '''

    @classmethod
    def yamale_dependencies(cls) -> set:
        return {NotesToCrowdAnki, NoteModelsToCrowdAnki, MediaGroupToCrowdAnki}

    @dataclass
    class Representation(RepresentationBase):
        folder: str
        notes: dict
        note_models: dict
        headers: dict
        media: Optional[dict] = field(default_factory=lambda: dict())

    @classmethod
    def from_repr(cls, data: Union[Representation, dict]):
        rep: cls.Representation = data if isinstance(data, cls.Representation) else cls.Representation.from_dict(data)
        return cls(
            rep=rep,
            crowd_anki_export=CrowdAnkiExport.create_or_get(rep.folder),
            notes_transform=NotesToCrowdAnki.from_repr(rep.notes),
            note_model_transform=NoteModelsToCrowdAnki.from_repr(rep.note_models),
            headers_transform=HeadersToCrowdAnki.from_repr(rep.headers),
            media_transform=MediaGroupToCrowdAnki.from_repr(rep.media) if rep.media else None
        )

    rep: Representation
    crowd_anki_export: CrowdAnkiExport
    notes_transform: NotesToCrowdAnki
    note_model_transform: NoteModelsToCrowdAnki
    headers_transform: HeadersToCrowdAnki
    media_transform: Optional[MediaGroupToCrowdAnki]

    def execute(self):
        headers = self.headers_transform.execute()
        ca_wrapper = CrowdAnkiJsonWrapper(headers)

        note_models: List[dict] = self.note_model_transform.execute()

        nm_name_to_id: dict = {model.part_id: model.part.id for model in self.note_model_transform.note_models}
        notes = self.notes_transform.execute(nm_name_to_id)

        media_files: Set[MediaFile] = set()
        if self.media_transform:
            media_files = self.media_transform.execute(self.crowd_anki_export.media_loc)

        ca_wrapper.media_files = sorted([m.filename for m in media_files])
        ca_wrapper.name = self.headers_transform.headers.name
        ca_wrapper.note_models = note_models
        ca_wrapper.notes = notes

        # Set to CrowdAnkiExport
        self.crowd_anki_export.write_to_files(ca_wrapper.data)



from dataclasses import dataclass
from typing import Union

from brain_brew.build_tasks.crowd_anki.headers_from_crowdanki import headers_default_values
from brain_brew.configuration.part_holder import PartHolder
from brain_brew.configuration.representation_base import RepresentationBase
from brain_brew.representation.yaml.headers import Headers


@dataclass
class HeadersToCrowdAnki:
    @dataclass
    class Representation(RepresentationBase):
        part_id: str

    @classmethod
    def from_repr(cls, data: Union[Representation, dict, str]):
        rep: cls.Representation
        if isinstance(data, cls.Representation):
            rep = data
        elif isinstance(data, dict):
            rep = cls.Representation.from_dict(data)
        else:
            rep = cls.Representation(part_id=data)  # Support single string being passed in

        return cls(
            rep=rep,
            headers=PartHolder.from_file_manager(rep.part_id).part
        )

    rep: Representation
    headers: Headers

    def execute(self) -> dict:
        headers = self.headers_to_crowd_anki(self.headers.data_without_name)

        return headers

    @staticmethod
    def headers_to_crowd_anki(headers_data: dict):
        return {**headers_default_values, **headers_data}




import logging

from brain_brew.commands.argument_reader import BBArgumentReader
# sys.path.append(os.path.join(os.path.dirname(__file__), "dist"))
# sys.path.append(os.path.dirname(__file__))
from brain_brew.configuration.file_manager import FileManager


def main():
    logging.basicConfig(level=logging.DEBUG)

    # Read in Arguments
    argument_reader = BBArgumentReader()
    command = argument_reader.get_parsed()

    # Create Singleton FileManager
    FileManager()

    command.execute()


if __name__ == "__main__":
    main()



import logging
from dataclasses import dataclass
from typing import List, Dict

from brain_brew.configuration.representation_base import RepresentationBase
from brain_brew.transformers.file_mapping import FileMapping
from brain_brew.transformers.note_model_mapping import NoteModelMapping


@dataclass
class SharedBaseCsvs:
    @dataclass(init=False)
    class Representation(RepresentationBase):
        file_mappings: List[FileMapping.Representation]
        note_model_mappings: List[NoteModelMapping.Representation]

        def __init__(self, file_mappings, note_model_mappings):
            self.file_mappings = list(map(FileMapping.Representation.from_dict, file_mappings))
            self.note_model_mappings = list(map(NoteModelMapping.Representation.from_dict, note_model_mappings))

        def get_file_mappings(self) -> List[FileMapping]:
            return list(map(FileMapping.from_repr, self.file_mappings))

    file_mappings: List[FileMapping]
    note_model_mappings: Dict[str, NoteModelMapping]

    @classmethod
    def map_nmm(cls, nmm_to_map):
        nmm = NoteModelMapping.from_repr(nmm_to_map)
        return nmm.get_note_model_mapping_dict()

    def verify_contents(self):
        errors = []

        for nm in self.note_model_mappings.values():
            try:
                nm.verify_contents()
            except KeyError as e:
                errors.append(e)

        # Check all referenced note models have a mapping
        for csv_map in self.file_mappings:
            for nm in csv_map.get_used_note_model_names():
                if nm not in self.note_model_mappings.keys():
                    errors.append(f"Missing Note Model Map for {nm}")

        # Check each of the Csvs (or their derivatives) contain all the necessary columns for their stated note model
        for cfm in self.file_mappings:
            note_model_names = cfm.get_used_note_model_names()
            available_columns = cfm.get_available_columns()

            referenced_note_models_maps = [value for key, value in self.note_model_mappings.items()
                                           if key in note_model_names]
            for nm_map in referenced_note_models_maps:
                for holder in nm_map.note_models.values():
                    if holder.part.name in note_model_names:
                        missing_columns = [col for col in holder.part.field_names_lowercase if
                                           col not in nm_map.csv_headers_map_to_note_fields(available_columns)]
                        if missing_columns:
                            logging.warning(f"Csvs are missing columns from {holder.part_id} {missing_columns}")

        if errors:
            raise Exception(errors)



import os
from dataclasses import dataclass
from typing import List

from brain_brew.build_tasks.crowd_anki.crowd_anki_generate import CrowdAnkiGenerate
from brain_brew.build_tasks.crowd_anki.headers_from_crowdanki import HeadersFromCrowdAnki
from brain_brew.build_tasks.crowd_anki.headers_to_crowd_anki import HeadersToCrowdAnki
from brain_brew.build_tasks.crowd_anki.media_group_from_crowd_anki import MediaGroupFromCrowdAnki
from brain_brew.build_tasks.crowd_anki.media_group_to_crowd_anki import MediaGroupToCrowdAnki
from brain_brew.build_tasks.crowd_anki.note_models_all_from_crowd_anki import NoteModelsAllFromCrowdAnki
from brain_brew.build_tasks.crowd_anki.note_models_to_crowd_anki import NoteModelsToCrowdAnki
from brain_brew.build_tasks.crowd_anki.notes_from_crowd_anki import NotesFromCrowdAnki
from brain_brew.build_tasks.crowd_anki.notes_to_crowd_anki import NotesToCrowdAnki
from brain_brew.build_tasks.csvs.csvs_generate import CsvsGenerate
from brain_brew.build_tasks.csvs.generate_guids_in_csvs import GenerateGuidsInCsvs
from brain_brew.build_tasks.csvs.notes_from_csvs import NotesFromCsvs
from brain_brew.build_tasks.deck_parts.headers_from_yaml_part import HeadersFromYamlPart
from brain_brew.build_tasks.deck_parts.media_group_from_folder import MediaGroupFromFolder
from brain_brew.build_tasks.deck_parts.note_model_from_yaml_part import NoteModelsFromYamlPart
from brain_brew.build_tasks.deck_parts.save_media_group_to_folder import SaveMediaGroupsToFolder
from brain_brew.build_tasks.deck_parts.save_note_models_to_folder import SaveNoteModelsToFolder
from brain_brew.commands.run_recipe.build_task import TopLevelBuildTask, BuildPartTask
from brain_brew.commands.run_recipe.parts_builder import PartsBuilder
from brain_brew.commands.run_recipe.top_level_builder import TopLevelBuilder
from brain_brew.interfaces.command import Command
from brain_brew.representation.generic.csv_file import CsvFile
from brain_brew.representation.yaml.note_model import NoteModel
from brain_brew.representation.yaml.yaml_object import YamlObject
from brain_brew.transformers.file_mapping import FileMapping
from brain_brew.transformers.note_model_mapping import NoteModelMapping
from brain_brew.utils import create_path_if_not_exists, filename_from_full_path, folder_name_from_full_path

RECIPE_MEDIA = "deck_media"
RECIPE_HEADERS = "deck_headers"
RECIPE_NOTES = "deck_notes"

LOC_RECIPES = "recipes/"
LOC_BUILD = "build/"
LOC_DATA = "src/data/"
LOC_HEADERS = "src/headers/"
LOC_NOTE_MODELS = "src/note_models/"
LOC_MEDIA = "src/media/"


@dataclass
class InitRepo(Command):
    crowdanki_folder: str
    delimiter: str

    def execute(self):
        self.setup_repo_structure()

        # Create the Deck Parts used
        headers_ca, note_models_all_ca, notes_ca, media_group_ca = self.parts_from_crowdanki(self.crowdanki_folder)

        headers = headers_ca.execute().part
        headers_name = LOC_HEADERS + "header1.yaml"
        headers.dump_to_yaml(headers_name)
        # TODO: desc file

        note_models = [m.part for m in note_models_all_ca.execute()]

        notes = notes_ca.execute().part
        used_note_models_in_notes = notes.get_all_known_note_model_names()

        media_group_ca.execute()

        note_model_mappings = [NoteModelMapping.Representation([model.name for model in note_models])]
        file_mappings: List[FileMapping.Representation] = []

        csv_files = []

        for model in note_models:
            if model.name in used_note_models_in_notes:
                csv_file_path = os.path.join(LOC_DATA, CsvFile.to_filename_csv(model.name, self.delimiter))
                column_headers = ["guid"] + model.field_names_lowercase + ["tags"]
                CsvFile.create_file_with_headers(csv_file_path, column_headers, delimiter=self.delimiter)

                file_mappings.append(FileMapping.Representation(
                    file=csv_file_path,
                    note_model=model.name,
                    delimiter=self.delimiter
                ))

                csv_files.append(csv_file_path)

        deck_path = os.path.join(LOC_BUILD, folder_name_from_full_path(self.crowdanki_folder))

        # Generate the Source files that will be kept in the repo
        save_note_models_to_folder = SaveNoteModelsToFolder.from_repr(SaveNoteModelsToFolder.Representation(
            [m.name for m in note_models], LOC_NOTE_MODELS, True
        ))
        model_name_to_file_dict = save_note_models_to_folder.execute()

        save_media_to_folder = SaveMediaGroupsToFolder.from_repr(SaveMediaGroupsToFolder.Representation(
            parts=[RECIPE_MEDIA], folder=LOC_MEDIA, recursive=True, clear_folder=True
        ))
        save_media_to_folder.execute()

        generate_csvs = CsvsGenerate.from_repr({
            'notes': RECIPE_NOTES,
            'note_model_mappings': note_model_mappings,
            'file_mappings': file_mappings
        })
        generate_csvs.execute()

        # Create Recipes

        # Anki to Source
        headers_recipe, note_models_all_recipe, notes_recipe, media_group_recipe = self.parts_from_crowdanki(deck_path)

        build_part_tasks: List[BuildPartTask] = [
            headers_recipe,
            notes_recipe,
            note_models_all_recipe,
            media_group_recipe,
        ]
        dp_builder = PartsBuilder(build_part_tasks)

        top_level_tasks: List[TopLevelBuildTask] = [dp_builder, save_media_to_folder, generate_csvs]
        self.create_yaml_from_top_level(top_level_tasks, os.path.join(LOC_RECIPES, "anki_to_source"))

        # Source to Anki
        note_models_from_yaml = [
            NoteModelsFromYamlPart.from_repr(NoteModelsFromYamlPart.Representation(name, file))
            for name, file in model_name_to_file_dict.items()
        ]

        media_group_from_folder = MediaGroupFromFolder.from_repr(MediaGroupFromFolder.Representation(
            part_id=RECIPE_MEDIA, source=LOC_MEDIA, recursive=True
        ))

        headers_from_yaml = HeadersFromYamlPart.from_repr(HeadersFromYamlPart.Representation(
            part_id=RECIPE_HEADERS, file=headers_name
        ))

        notes_from_csv = NotesFromCsvs.from_repr({
            'part_id': RECIPE_NOTES,
            'note_model_mappings': note_model_mappings,
            'file_mappings': file_mappings
        })

        build_part_tasks: List[BuildPartTask] = note_models_from_yaml + [
            headers_from_yaml,
            notes_from_csv,
            media_group_from_folder,
        ]
        dp_builder = PartsBuilder(build_part_tasks)

        generate_guids_in_csv = GenerateGuidsInCsvs.from_repr(GenerateGuidsInCsvs.Representation(
            source=csv_files, columns=["guid"], delimiter=self.delimiter
        ))

        generate_crowdanki = CrowdAnkiGenerate.from_repr(CrowdAnkiGenerate.Representation(
            folder=deck_path,
            notes=NotesToCrowdAnki.Representation(
                part_id=RECIPE_NOTES
            ).encode(),
            headers=RECIPE_HEADERS,
            media=MediaGroupToCrowdAnki.Representation(
                parts=[RECIPE_MEDIA]
            ).encode(),
            note_models=NoteModelsToCrowdAnki.Representation(
                parts=[NoteModelsToCrowdAnki.NoteModelListItem.Representation(name).encode()
                       for name, file in model_name_to_file_dict.items()]
            ).encode()
        ))

        top_level_tasks: List[TopLevelBuildTask] = [generate_guids_in_csv, dp_builder, generate_crowdanki]
        source_to_anki_path = os.path.join(LOC_RECIPES, "source_to_anki.yaml")
        self.create_yaml_from_top_level(top_level_tasks, source_to_anki_path)

        print(f"\nRepo Init complete. You should now run `brainbrew run {source_to_anki_path}`")

    @staticmethod
    def create_yaml_from_top_level(top_tasks: List[TopLevelBuildTask], filepath: str):
        tl_builder = TopLevelBuilder(top_tasks)

        encoded_top_level_tasks = tl_builder.encode()
        # print(encoded_top_level_tasks)

        model_yaml_file_name = YamlObject.to_filename_yaml(filepath)
        YamlObject.dump_to_yaml_file(model_yaml_file_name, encoded_top_level_tasks)

    @staticmethod
    def parts_from_crowdanki(folder: str):
        headers_ca = HeadersFromCrowdAnki.from_repr(HeadersFromCrowdAnki.Representation(
            source=folder, part_id=RECIPE_HEADERS
        ))
        note_models_all_ca = NoteModelsAllFromCrowdAnki.from_repr(NoteModelsAllFromCrowdAnki.Representation(
            source=folder
        ))
        notes_ca = NotesFromCrowdAnki.from_repr(NotesFromCrowdAnki.Representation(
            source=folder, part_id=RECIPE_NOTES
        ))
        media_group_ca = MediaGroupFromCrowdAnki.from_repr(MediaGroupFromFolder.Representation(
            source=folder, part_id=RECIPE_MEDIA
        ))
        return headers_ca, note_models_all_ca, notes_ca, media_group_ca

    @staticmethod
    def setup_repo_structure():
        create_path_if_not_exists(LOC_RECIPES)
        create_path_if_not_exists(LOC_BUILD)
        create_path_if_not_exists(LOC_DATA)
        create_path_if_not_exists(LOC_HEADERS)
        create_path_if_not_exists(LOC_NOTE_MODELS)
        create_path_if_not_exists(LOC_MEDIA)



from brain_brew.configuration.file_manager import FileManager


def get_new_file_manager():
    FileManager.clear_instance()
    return FileManager()


# class TestSingletonConstructor:
#     def test_runs(self, global_config):
#         fm = get_new_file_manager()
#         assert isinstance(fm, FileManager)
#
#     def test_returns_existing_singleton(self):
#         fm = get_new_file_manager()
#         fm.known_files_dict = {'test': None}
#         fm2 = FileManager.get_instance()
#
#         assert fm2.known_files_dict == {'test': None}
#         assert fm2 == fm
#
#     def test_raises_error(self):
#         with pytest.raises(Exception):
#             FileManager()
#             FileManager()
#
#
# class TestFindMediaFiles:
#     def test_finds(self):
#         fm = get_new_file_manager()
#
#         assert len(fm.known_media_files_dict) == 2



from dataclasses import dataclass, field
from typing import Optional, Union

from brain_brew.configuration.representation_base import RepresentationBase
from brain_brew.interfaces.yamale_verifyable import YamlRepr
from brain_brew.representation.generic.html_file import HTMLFile
from brain_brew.representation.yaml.headers import Headers


@dataclass
class HeadersOverride(YamlRepr):
    @classmethod
    def task_name(cls) -> str:
        return r"headers_override"

    @classmethod
    def yamale_schema(cls) -> str:
        return f'''\
            crowdanki_uuid: str(required=False)
            deck_description_html_file: str(required=False)
            name: str(required=False)
        '''

    @dataclass
    class Representation(RepresentationBase):
        crowdanki_uuid: Optional[str] = field(default=None)
        deck_description_html_file: Optional[str] = field(default=None)
        name: Optional[str] = field(default=None)

    @classmethod
    def from_repr(cls, data: Union[Representation, dict]):
        rep: cls.Representation = data if isinstance(data, cls.Representation) else cls.Representation.from_dict(data)
        return cls(
            rep=rep,
            crowdanki_uuid=rep.crowdanki_uuid,
            deck_desc_html_file=HTMLFile.create_or_get(rep.deck_description_html_file),
            name=rep.name
        )

    rep: Representation
    crowdanki_uuid: Optional[str]
    deck_desc_html_file: Optional[HTMLFile]
    name: Optional[str]

    def override(self, header: Headers):
        if self.deck_desc_html_file:
            header.description = self.deck_desc_html_file.get_data(deep_copy=True)

        if self.crowdanki_uuid:
            header.crowdanki_uuid = self.crowdanki_uuid

        if self.name:
            header.name = self.name

        return header



import json

_encoding = "utf-8"


class JsonFile:
    @staticmethod
    def pretty_print(data):
        return json.dumps(data, indent=4)

    @staticmethod
    def to_filename_json(filename: str):
        if filename[-5:] != ".json":
            return filename + ".json"
        return filename

    @staticmethod
    def read_file(file_location):
        with open(JsonFile.to_filename_json(file_location), "r", encoding=_encoding) as read_file:
            return json.load(read_file)

    @staticmethod
    def write_file(file_location, data):
        with open(JsonFile.to_filename_json(file_location), "w+", encoding=_encoding) as write_file:
            json.dump(data, write_file, indent=4, sort_keys=False, ensure_ascii=False)



from argparse import ArgumentParser, ArgumentError
from unittest.mock import patch

import pytest

from brain_brew.commands.argument_reader import BBArgumentReader, Commands


@pytest.fixture()
def arg_reader_test1():
    return BBArgumentReader(test_mode=True)


def test_constructor(arg_reader_test1):
    assert isinstance(arg_reader_test1, BBArgumentReader)
    assert isinstance(arg_reader_test1, ArgumentParser)


class TestArguments:
    class CommandRun:
        @pytest.mark.parametrize("arguments", [
            ([Commands.RUN_RECIPE.value]),
            ([Commands.RUN_RECIPE.value, ""]),
        ])
        def test_broken_arguments(self, arg_reader_test1, arguments):
            def raise_exit(message):
                raise SystemExit

            with pytest.raises(SystemExit):
                with patch.object(BBArgumentReader, "error", side_effect=raise_exit):
                    arg_reader_test1.get_parsed(arguments)

        @pytest.mark.parametrize("arguments, recipe, verify_only", [
            ([Commands.RUN_RECIPE.value, "test_recipe.yaml"], "test_recipe.yaml", False),
            ([Commands.RUN_RECIPE.value, "test_recipe.yaml", "--verify"], "test_recipe.yaml", True),
            ([Commands.RUN_RECIPE.value, "test_recipe.yaml", "-v"], "test_recipe.yaml", True),
        ])
        def test_correct_arguments(self, arg_reader_test1, arguments, recipe, verify_only):
            parsed_args = arg_reader_test1.parse_args(arguments)

            assert parsed_args.recipe == recipe
            assert parsed_args.verify_only == verify_only

    class CommandInit:
        @pytest.mark.parametrize("arguments, location", [
            (["init", "crowdankifolder72"], "crowdankifolder72"),
        ])
        def test_correct_arguments(self, arg_reader_test1, arguments, location):
            parsed_args = arg_reader_test1.parse_args(arguments)

            assert parsed_args.crowdanki_folder == location



from dataclasses import dataclass, field
from typing import Optional, Union, List

from brain_brew.commands.run_recipe.build_task import BuildPartTask
from brain_brew.configuration.part_holder import PartHolder
from brain_brew.configuration.representation_base import RepresentationBase
from brain_brew.representation.yaml.media_group import MediaGroup
from brain_brew.transformers.create_media_group_from_location import create_media_group_from_location


@dataclass
class MediaGroupFromFolder(BuildPartTask):
    @classmethod
    def task_name(cls) -> str:
        return r"media_group_from_folder"

    @classmethod
    def yamale_schema(cls) -> str:
        return f'''\
            part_id: str()
            source: str()
            save_to_file: str(required=False)
            recursive: bool(required=False)
            filter_whitelist_from_parts: list(str(), required=False)
            filter_blacklist_from_parts: list(str(), required=False)
        '''

    @dataclass
    class Representation(RepresentationBase):
        part_id: str
        source: str
        filter_blacklist_from_parts: List[str] = field(default_factory=list)
        filter_whitelist_from_parts: List[str] = field(default_factory=list)
        recursive: Optional[bool] = field(default=True)
        save_to_file: Optional[str] = field(default=None)

    @classmethod
    def from_repr(cls, data: Union[Representation, dict]):
        rep: cls.Representation = data if isinstance(data, cls.Representation) else cls.Representation.from_dict(data)
        return cls(
            rep=rep,
            part=create_media_group_from_location(
                part_id=rep.part_id,
                save_to_file=rep.save_to_file,
                media_group=MediaGroup.from_directory(rep.source, rep.recursive),
                groups_to_blacklist=list(holder.part for holder in
                                         map(PartHolder.from_file_manager, rep.filter_blacklist_from_parts)),
                groups_to_whitelist=list(holder.part for holder in
                                         map(PartHolder.from_file_manager, rep.filter_whitelist_from_parts))
                # match criteria
            )
        )

    rep: Representation
    part: MediaGroup

    def execute(self):
        pass



from dataclasses import dataclass

from brain_brew.representation.json.wrappers_for_crowd_anki import CA_NAME, CA_DESCRIPTION, CA_UUID
from brain_brew.representation.yaml.yaml_object import YamlObject


@dataclass
class Headers(YamlObject):
    data: dict

    @classmethod
    def from_yaml_file(cls, filename: str):
        return cls(data=cls.read_to_dict(filename))

    def encode(self) -> dict:
        return self.data

    @property
    def name(self) -> str:
        return self.data[CA_NAME]

    @name.setter
    def name(self, desc: str):
        self.data[CA_NAME] = desc

    @property
    def description(self) -> str:
        return self.data.get(CA_DESCRIPTION, "")

    @description.setter
    def description(self, desc: str):
        self.data[CA_DESCRIPTION] = desc

    @property
    def crowdanki_uuid(self) -> str:
        return self.data.get(CA_UUID, "")

    @crowdanki_uuid.setter
    def crowdanki_uuid(self, desc: str):
        self.data[CA_UUID] = desc

    @property
    def data_without_name(self) -> dict:
        return {k: v for k, v in sorted(self.data.items()) if k != CA_NAME}



import logging
from dataclasses import dataclass, field
from typing import List, Union, Optional

from brain_brew.commands.run_recipe.build_task import TopLevelBuildTask
from brain_brew.configuration.representation_base import RepresentationBase
from brain_brew.representation.generic.csv_file import CsvFile
from brain_brew.utils import single_item_to_list, generate_anki_guid


@dataclass
class GenerateGuidsInCsvs(TopLevelBuildTask):
    execute_immediately = True

    @classmethod
    def task_name(cls) -> str:
        return r'generate_guids_in_csvs'

    @classmethod
    def task_regex(cls) -> str:
        return r'generate_guids_in_csv[s]?'

    @classmethod
    def yamale_schema(cls) -> str:
        return f'''\
            source: any(str(), list(str()))
            columns: any(str(), list(str()))
            delimiter: str(required=False)
        '''

    @dataclass
    class Representation(RepresentationBase):
        source: Union[str, List[str]]
        columns: Union[str, List[str]]
        delimiter: Optional[str] = field(default=None)

        def encode_filter(self, key, value):
            if not super().encode_filter(key, value):
                return False
            if key == 'delimiter' and all(CsvFile.delimiter_matches_file_type(value, f) for f in single_item_to_list(self.source)):
                return False
            return True

    @classmethod
    def from_repr(cls, data: Union[Representation, dict]):
        rep: cls.Representation = data if isinstance(data, cls.Representation) else cls.Representation.from_dict(data)
        csv_files = [CsvFile.create_or_get(csv) for csv in single_item_to_list(rep.source)]
        for c in csv_files:
            c.set_delimiter(rep.delimiter)
            c.read_file()
        return cls(
            rep=rep,
            sources=csv_files,
            columns=rep.columns
        )

    rep: Representation
    sources: List[CsvFile]
    columns: List[str]

    def execute(self):
        logging.info("Attempting to generate Guids")

        errors = []

        # Make sure the columns exist on all
        for source in self.sources:
            missing = [c for c in self.columns if c not in source.column_headers]
            if any(missing):
                errors.append(f"Csv '{source.file_location}' does not contain all the specified columns: {missing}")

        if errors:
            raise KeyError(errors)

        for source in self.sources:
            guids_generated = 0
            data = source.get_data()
            for row in data:
                for column_name in row.keys():
                    if column_name in self.columns and not row[column_name]:
                        row[column_name] = generate_anki_guid()
                        guids_generated += 1
            if guids_generated > 0:
                logging.info(f"Generated {guids_generated} guids in csv '{source.file_location}'")
                source.set_data(data)
                source.write_file()

        logging.info("Generate guids complete")



from enum import Enum

import sys
from argparse import ArgumentParser

from brain_brew.front_matter import latest_version_number
from brain_brew.commands.init_repo.init_repo import InitRepo
from brain_brew.commands.run_recipe.run_recipe import RunRecipe
from brain_brew.interfaces.command import Command


class Commands(Enum):
    RUN_RECIPE = "run"
    INIT_REPO = "init"


class BBArgumentReader(ArgumentParser):
    def __init__(self, test_mode=False):
        super().__init__(
            prog="brainbrew",
            description='Manage Flashcards by transforming them to various types.'
        )

        self._set_parser_arguments()

        if not test_mode and len(sys.argv) == 1:
            self.print_help(sys.stderr)
            sys.exit(1)

    def _set_parser_arguments(self):

        subparsers = self.add_subparsers(parser_class=ArgumentParser, help='Commands that can be run', dest="command")

        parser_run = subparsers.add_parser(
            Commands.RUN_RECIPE.value,
            help="Run a recipe file. This will convert some data to another format, based on the instructions in the recipe file."
        )
        parser_run.add_argument(
            "recipe",
            metavar="recipe",
            type=str,
            help="Yaml file to use as the recipe"
        )
        parser_run.add_argument(
            "--verify", "-v",
            action="store_true",
            dest="verify_only",
            default=False,
            help="Only verify the recipe contents, without running it."
        )

        parser_init = subparsers.add_parser(
            Commands.INIT_REPO.value,
            help="Initialise a Brain Brew repository, using a CrowdAnki export as the base data."
        )
        parser_init.add_argument(
            "crowdanki_folder",
            metavar="crowdanki_folder",
            type=str,
            help="The folder that stores the CrowdAnki files to build this repo from"
        )
        parser_init.add_argument(
            '--delimiter',
            dest='delimiter',
            action='store',
            help="Set the delimiter for Csv files to specific character",
            type=str
        )
        parser_init.add_argument(
            "--delimitertab", "--tab",
            action="store_true",
            dest="delimiter_tab",
            default=False,
            help="Use tabs as the delimiter for Csv files"
        )

    def get_parsed(self, override_args=None) -> Command:
        parsed_args = self.parse_args(args=override_args)

        if parsed_args.command == Commands.RUN_RECIPE.value:
            # Required
            recipe = self.error_if_blank(parsed_args.recipe)

            # Optional
            verify_only = parsed_args.verify_only

            return RunRecipe(
                recipe_file_name=recipe,
                verify_only=verify_only
            )

        if parsed_args.command == Commands.INIT_REPO.value:
            # Required
            crowdanki_folder = parsed_args.crowdanki_folder
            delimiter = parsed_args.delimiter
            delimiter_tab = parsed_args.delimiter_tab

            return InitRepo(
                crowdanki_folder=crowdanki_folder,
                delimiter="\t" if delimiter_tab else delimiter
            )

        raise KeyError("Unknown Command")

    def error_if_blank(self, arg):
        if arg == "" or arg is None:
            self.error("Required argument missing")
        return arg

    def error(self, message):
        sys.stderr.write('error: %s\n' % message)
        self.print_help()
        sys.exit(2)

    def print_help(self, message=None):
        print(f"Brain Brew v{latest_version_number()}")
        super().print_help(message)



from typing import List


CA_NOTE_MODELS = "note_models"
CA_NOTES = "notes"
CA_MEDIA_FILES = "media_files"
CA_CHILDREN = "children"
CA_TYPE = "__type__"
CA_NAME = "name"
CA_DESCRIPTION = "desc"
CA_UUID = "crowdanki_uuid"

NOTE_MODEL = "note_model_uuid"
FLAGS = "flags"
GUID = "guid"
TAGS = "tags"
FIELDS = "fields"


class CrowdAnkiJsonWrapper:
    data: dict

    def __init__(self, data: dict = None):
        self.data = data

    @property
    def children(self) -> list:
        return self.data.get(CA_CHILDREN, [])

    @property
    def note_models(self) -> list:
        return CrowdAnkiJsonWrapper.get_from_self_and_children_recursively(self.data, [], CA_NOTE_MODELS)


    @note_models.setter
    def note_models(self, value: list):
        self.data[CA_NOTE_MODELS] = value

    @property
    def notes(self) -> list:
        return CrowdAnkiJsonWrapper.get_from_self_and_children_recursively(self.data, [], CA_NOTES)

    @notes.setter
    def notes(self, value: list):
        self.data[CA_NOTES] = value

    @property
    def media_files(self) -> list:
        return CrowdAnkiJsonWrapper.get_from_self_and_children_recursively(self.data, [], CA_MEDIA_FILES)

    @media_files.setter
    def media_files(self, value: list):
        self.data[CA_MEDIA_FILES] = value

    @property
    def name(self) -> list:
        return self.data.get(CA_NAME, [])

    @name.setter
    def name(self, value: list):
        self.data[CA_NAME] = value

    @staticmethod
    def get_from_self_and_children_recursively(data: dict, running_data: list, key_name: str):
        running_data += data.get(key_name, [])
        children = data.get(CA_CHILDREN, [])
        if isinstance(children, list):
            for child in children:
                running_data = CrowdAnkiJsonWrapper.get_from_self_and_children_recursively(child, running_data, key_name)
        return running_data


class CrowdAnkiNoteWrapper:
    data: dict

    def __init__(self, data: dict = None):
        self.data = data

    @property
    def note_model(self) -> str:
        return self.data.get(NOTE_MODEL)

    @note_model.setter
    def note_model(self, value: str):
        self.data[NOTE_MODEL] = value

    @property
    def flags(self) -> int:
        return self.data.get(FLAGS)

    @flags.setter
    def flags(self, value: int):
        self.data[FLAGS] = value

    @property
    def guid(self) -> str:
        return self.data.get(GUID)

    @guid.setter
    def guid(self, value: str):
        self.data[GUID] = value

    @property
    def tags(self) -> list:
        return self.data.get(TAGS, [])

    @tags.setter
    def tags(self, value: list):
        self.data[TAGS] = value

    @property
    def fields(self) -> List[str]:
        return self.data.get(FIELDS, [])

    @fields.setter
    def fields(self, value: List[str]):
        self.data[FIELDS] = value



import logging
import os
import random
import re
import shutil
import string
from pathlib import Path
from typing import List


def blank_str_if_none(s):
    return '' if s is None else s


def list_of_str_to_lowercase(list_of_strings):
    return [entry.lower() for entry in list_of_strings]


def single_item_to_list(item):
    if isinstance(item, list):
        return item
    if item is None:
        return []
    return [item]


def str_to_lowercase_no_separators(str_to_tidy: str):
    return re.sub(r'[\s+_-]+', '', str_to_tidy.lower())


def filename_from_full_path(full_path):
    return re.findall(r'[^\\/:*?"<>|\r\n]+$', full_path)[0]


def folder_name_from_full_path(full_path):
    return re.findall(r'[^\\/:*?"<>|\r\n]+[/]?$', full_path)[0]


def split_by_regex(str_to_split: str, pattern: str) -> List[str]:
    return re.split(pattern, str_to_split)


def find_media_in_field(field_value: str) -> List[str]:
    if not field_value:
        return []

    images = re.findall(r'<\s*?img.*?src="(.*?)"[^>]*?>', field_value)
    audio = re.findall(r'\[sound:(.*?)]', field_value)

    return images + audio


def find_all_files_in_directory(directory, recursive=False):
    found_files = []
    for path, dirs, files in os.walk(directory):
        for file in files:
            found_files.append(os.path.join(path, file))
        if not recursive:
            return found_files
    return found_files


def create_path_if_not_exists(path):
    dir_name = os.path.dirname(path)
    if not Path(dir_name).is_dir():
        logging.warning(f"Creating missing filepath '{dir_name}'")
        os.makedirs(dir_name, exist_ok=True)


def clear_contents_of_folder(path):
    for filename in os.listdir(path):
        file_path = os.path.join(path, filename)
        try:
            if os.path.isfile(file_path) or os.path.islink(file_path):
                os.unlink(file_path)
            elif os.path.isdir(file_path):
                shutil.rmtree(file_path)
        except Exception as e:
            print('Failed to delete %s. Reason: %s' % (file_path, e))


def split_tags(tags_value: str) -> list:
    split = [entry.strip() for entry in re.split(r';\s*|,\s*|\s+', tags_value)]
    while "" in split:
        split.remove("")
    return split


def join_tags(tags_list: list) -> str:
    return ", ".join(tags_list)  # TODO: Make configurable


def generate_anki_guid() -> str:
    """Return a base91-encoded 64bit random number."""

    def base62(num: int, extra: str = "") -> str:
        s = string
        table = s.ascii_letters + s.digits + extra
        buf = ""
        while num:
            num, i = divmod(num, len(table))
            buf = table[i] + buf
        return buf

    _base91_extra_chars = "!#$%&()*+,-./:;<=>?@[]^_`{|}~"

    def base91(num: int) -> str:
        # all printable characters minus quotes, backslash and separators
        return base62(num, _base91_extra_chars)

    return base91(random.randint(0, 2 ** 64 - 1))


def sort_dict(data, sort_by_keys, reverse_sort, case_insensitive_sort):
    if sort_by_keys:
        if case_insensitive_sort:
            def sort_method(i):
                return tuple((i[column] == "", i[column].lower()) for column in sort_by_keys)
        else:
            def sort_method(i):
                return tuple((i[column] == "", i[column]) for column in sort_by_keys)

        return sorted(data, key=sort_method, reverse=reverse_sort)
    elif reverse_sort:
        return list(reversed(data))

    return data



import logging
import os
from typing import List

from brain_brew.representation.generic.html_file import HTMLFile
from brain_brew.representation.yaml.note_model import NoteModel, CSS_FILE, TEMPLATES
from brain_brew.representation.yaml.note_model_template import HTML_FILE as TEMPLATE_HTML_FILE, NAME as TEMPLATE_NAME, BROWSER_HTML_FILE as TEMPLATE_BROWSER_HTML_FILE
from brain_brew.representation.yaml.yaml_object import YamlObject
from brain_brew.utils import create_path_if_not_exists, clear_contents_of_folder


def save_note_model_to_location(
        model: NoteModel,
        folder: str,
        clear_folder: bool
) -> str:

    nm_folder = os.path.join(folder, model.name + '/')
    create_path_if_not_exists(nm_folder)

    if clear_folder:
        clear_contents_of_folder(nm_folder)

    model_encoded = model.encode_as_part_with_empty_file_references()

    model_encoded[CSS_FILE.name] = os.path.join(nm_folder, "style.css")
    HTMLFile.write_file(model_encoded[CSS_FILE.name], model.css)

    templates_dict = {t.name: t for t in model.templates}

    for template_data in model_encoded[TEMPLATES.name]:
        name = template_data[TEMPLATE_NAME.name]
        template = templates_dict[name]
        t_data, b_t_data = template.get_template_files_data()

        template_data[TEMPLATE_HTML_FILE.name] = os.path.join(nm_folder, HTMLFile.to_filename_html(name))
        HTMLFile.write_file(template_data[TEMPLATE_HTML_FILE.name], t_data)

        if TEMPLATE_BROWSER_HTML_FILE.name in template_data and b_t_data is not None:
            template_data[TEMPLATE_BROWSER_HTML_FILE.name] = os.path.join(nm_folder, HTMLFile.to_filename_html(name + "_browser"))
            HTMLFile.write_file(template_data[TEMPLATE_HTML_FILE.name], b_t_data)

    model_yaml_file_name = YamlObject.to_filename_yaml(os.path.join(nm_folder, model.name))
    YamlObject.dump_to_yaml_file(model_yaml_file_name, model_encoded)

    return model_yaml_file_name



from dataclasses import dataclass, field
from typing import Union, List

from brain_brew.configuration.part_holder import PartHolder
from brain_brew.configuration.representation_base import RepresentationBase
from brain_brew.interfaces.yamale_verifyable import YamlRepr
from brain_brew.representation.yaml.note_model import NoteModel


@dataclass
class NoteModelsToCrowdAnki(YamlRepr):
    @classmethod
    def task_name(cls) -> str:
        return r'note_models_to_crowd_anki'

    @classmethod
    def yamale_schema(cls) -> str:
        return f'''\
            parts: list(include('{cls.NoteModelListItem.task_name()}'))
        '''

    @classmethod
    def yamale_dependencies(cls) -> set:
        return {cls.NoteModelListItem}

    @dataclass
    class NoteModelListItem(YamlRepr):
        @classmethod
        def task_name(cls) -> str:
            return r'note_models_to_crowd_anki_item'

        @classmethod
        def yamale_schema(cls) -> str:
            return f'''\
                part_id: str()
            '''

        @dataclass
        class Representation(RepresentationBase):
            part_id: str
            # TODO: fields: Optional[List[str]]
            # TODO: templates: Optional[List[str]]

        @classmethod
        def from_repr(cls, data: Union[Representation, dict, str]):
            rep: cls.Representation
            if isinstance(data, cls.Representation):
                rep = data
            elif isinstance(data, dict):
                rep = cls.Representation.from_dict(data)
            else:
                rep = cls.Representation(part_id=data)  # Support string

            return cls(
                rep=rep,
                part_to_read=rep.part_id
            )

        def get_note_model(self) -> PartHolder[NoteModel]:
            self.part = PartHolder.from_file_manager(self.part_to_read)
            return self.part  # Todo: add filters in here

        rep: Representation
        part: PartHolder[NoteModel] = field(init=False)
        part_to_read: str

    @dataclass
    class Representation(RepresentationBase):
        parts: List[Union[dict, str]]

    @classmethod
    def from_repr(cls, data: Union[Representation, dict, List[str]]):
        rep: cls.Representation
        if isinstance(data, cls.Representation):
            rep = data
        elif isinstance(data, dict):
            rep = cls.Representation.from_dict(data)
        else:
            rep = cls.Representation(parts=data)  # Support list of Note Models

        note_model_items = list(map(cls.NoteModelListItem.from_repr, rep.parts))
        return cls(
            rep=rep,
            note_models=[nm.get_note_model() for nm in note_model_items]
        )

    rep: Representation
    note_models: List[PartHolder[NoteModel]]

    def execute(self) -> List[dict]:
        return [model.part.encode_as_crowdanki() for model in self.note_models]



from abc import ABC, abstractmethod
from pathlib import Path

from ruamel.yaml import YAML

from brain_brew.utils import create_path_if_not_exists

yaml_load = YAML(typ='safe')


yaml_dump = YAML()
yaml_dump.preserve_quotes = False
yaml_dump.indent(mapping=2, sequence=2, offset=0)
yaml_dump.representer.ignore_aliases = lambda *data: True
# yaml.sort_base_mapping_type_on_output = False


class YamlObject(ABC):
    @staticmethod
    def read_to_dict(filename: str):
        filename = YamlObject.to_filename_yaml(filename)

        if not Path(filename).is_file():
            raise FileNotFoundError(filename)

        with open(filename) as file:
            return yaml_load.load(file)

    @staticmethod
    def to_filename_yaml(filename: str):
        if filename[-5:] != ".yaml" and filename[-4:] != ".yml":
            return filename + ".yaml"
        return filename

    @abstractmethod
    def encode(self) -> dict:
        pass

    @classmethod
    @abstractmethod
    def from_yaml_file(cls, filename: str) -> 'YamlObject':
        pass

    def dump_to_yaml(self, filepath):
        self.dump_to_yaml_file(filepath, self.encode())

    @classmethod
    def dump_to_yaml_file(cls, filepath, data):
        filepath = YamlObject.to_filename_yaml(filepath)

        create_path_if_not_exists(filepath)

        with open(filepath, 'w') as fp:
            yaml_dump.dump(data, fp)




from textwrap import indent, dedent
from typing import Dict, Type, List, Set

from brain_brew.build_tasks.crowd_anki.crowd_anki_generate import CrowdAnkiGenerate
from brain_brew.build_tasks.csvs.csvs_generate import CsvsGenerate
from brain_brew.build_tasks.csvs.generate_guids_in_csvs import GenerateGuidsInCsvs
from brain_brew.build_tasks.deck_parts.save_media_group_to_folder import SaveMediaGroupsToFolder
from brain_brew.build_tasks.deck_parts.save_note_models_to_folder import SaveNoteModelsToFolder
from brain_brew.commands.run_recipe.build_task import BuildTask, TopLevelBuildTask
from brain_brew.commands.run_recipe.parts_builder import PartsBuilder
from brain_brew.commands.run_recipe.recipe_builder import RecipeBuilder
from brain_brew.interfaces.yamale_verifyable import YamlRepr


class TopLevelBuilder(YamlRepr, RecipeBuilder):
    @classmethod
    def known_task_dict(cls) -> Dict[str, Type[BuildTask]]:
        values = TopLevelBuildTask.get_all_task_regex(cls.yamale_dependencies())
        return values

    @classmethod
    def build_yamale(cls):
        separator = '\n---\n'
        top_level = cls.yamale_dependencies()

        builder: List[str] = [cls.build_yamale_root_node(top_level), separator]

        def to_sorted_yamale_string(lines: Set[Type[BuildTask]]):
            return [f'''{line.task_name()}:\n{indent(dedent(line.yamale_schema()), '    ')}'''
                    for line in sorted(lines, key=lambda x: x.task_name())]

        # Schema
        builder += to_sorted_yamale_string(top_level)

        builder.append(separator)

        # Dependencies
        def resolve_dependencies(deps: Set[Type[BuildTask]]) -> Set[Type[BuildTask]]:
            result = set()
            for d in deps:
                result.add(d)
                result = result.union(resolve_dependencies(d.yamale_dependencies()))
            return result

        children = resolve_dependencies(cls.yamale_dependencies())
        builder += to_sorted_yamale_string({dep for dep in children if dep not in top_level})

        return '\n'.join(builder)

    @classmethod
    def parse_and_read(cls, filename, verify_only: bool) -> 'TopLevelBuilder':
        recipe_data = cls.read_to_dict(filename)

        from brain_brew.configuration.yaml_verifier import YamlVerifier
        YamlVerifier.get_instance().verify_recipe(filename)

        if verify_only:
            return None

        return cls.from_list(recipe_data)

    @classmethod
    def task_name(cls) -> str:
        pass

    @classmethod
    def yamale_schema(cls) -> str:
        pass

    @classmethod
    def from_repr(cls, data: dict):
        pass

    def encode(self) -> list:
        return self.tasks_to_encoded()

    @classmethod
    def from_yaml_file(cls, filename: str):
        pass

    @classmethod
    def yamale_dependencies(cls) -> Set[Type[TopLevelBuildTask]]:
        return {
            PartsBuilder,
            CrowdAnkiGenerate, CsvsGenerate,
            GenerateGuidsInCsvs, SaveMediaGroupsToFolder, SaveNoteModelsToFolder
        }



from textwrap import dedent
from typing import List, Set

import pytest

from brain_brew.representation.yaml.notes import Note, NoteGrouping, Notes, \
    NOTES, NOTE_GROUPINGS, FIELDS, GUID, NOTE_MODEL, TAGS, FLAGS

working_notes = {
    "test1": {FIELDS: ['first'], GUID: "12345", NOTE_MODEL: "model_name", TAGS: ['noun', 'other']},
    "test2": {FIELDS: ['english', 'german'], GUID: "sdfhfghsvsdv", NOTE_MODEL: "LL Test", TAGS: ['marked']},
    "no_note_model": {FIELDS: ['first'], GUID: "12345", TAGS: ['noun', 'other']},
    "no_note_model2": {FIELDS: ['second'], GUID: "67890", TAGS: ['noun', 'other']},
    "no_tags1": {FIELDS: ['first'], GUID: "12345", NOTE_MODEL: "model_name"},
    "no_tags2": {FIELDS: ['first'], GUID: "12345", NOTE_MODEL: "model_name", TAGS: []},
    "no_model_or_tags": {FIELDS: ['first'], GUID: "12345"},
    "test1_with_default_flags": {FIELDS: ['first'], GUID: "12345", NOTE_MODEL: "model_name", TAGS: ['noun', 'other'], FLAGS: 0},
    "test1_with_flags": {FIELDS: ['first'], GUID: "12345", NOTE_MODEL: "model_name", TAGS: ['noun', 'other'], FLAGS: 1},
}

working_note_groupings = {
    "nothing_grouped": {NOTES: [working_notes["test1"], working_notes["test2"]]},
    "note_model_grouped": {NOTES: [working_notes["no_note_model"], working_notes["no_note_model2"]], NOTE_MODEL: "model_name"},
    "note_model_grouped2": {NOTES: [working_notes["no_note_model"], working_notes["no_note_model2"]], NOTE_MODEL: "different_model"},
    "tags_grouped": {NOTES: [working_notes["no_tags1"], working_notes["no_tags2"]], TAGS: ["noun", "other"]},
    "tags_grouped_as_addition": {NOTES: [working_notes["test1"], working_notes["test2"]], TAGS: ["test", "recent"]},
    "model_and_tags_grouped": {NOTES: [working_notes["no_model_or_tags"], working_notes["no_model_or_tags"]], NOTE_MODEL: "model_name", TAGS: ["noun", "other"]}
}

working_dpns = {
    "one_group": {NOTE_GROUPINGS: [working_note_groupings["nothing_grouped"]]},
    "two_groups_two_models": {NOTE_GROUPINGS: [working_note_groupings["nothing_grouped"], working_note_groupings["note_model_grouped"]]},
    "two_groups_three_models": {NOTE_GROUPINGS: [working_note_groupings["nothing_grouped"], working_note_groupings["note_model_grouped2"]]},
}


@pytest.fixture(params=working_notes.values())
def note_fixtures(request):
    return Note.from_dict(request.param)


@pytest.fixture(params=working_note_groupings.values())
def note_grouping_fixtures(request):
    return NoteGrouping.from_dict(request.param)


class TestConstructor:
    class TestNote:
        @pytest.mark.parametrize("fields, guid, note_model, tags, flags, media", [
            ([], "", "", [], 0, {}),
            (None, None, None, None, None, None),
            (["test", "blah", "whatever"], "1234567890x", "model_name", ["noun"], 1, {}),
            (["test", "blah", "<img src=\"animal.jpg\">"], "1234567890x", "model_name", ["noun"], 2, {"animal.jpg"}),
        ])
        def test_constructor(self, fields: List[str], guid: str, note_model: str, tags: List[str], flags: int, media: Set[str]):
            note = Note(fields=fields, guid=guid, note_model=note_model, tags=tags, flags=flags)

            assert isinstance(note, Note)
            assert note.fields == fields
            assert note.guid == guid
            assert note.note_model == note_model
            assert note.tags == tags
            assert note.flags == flags
            # assert note.media_references == media

        def test_from_dict(self, note_fixtures):
            assert isinstance(note_fixtures, Note)

    class TestNoteGrouping:
        def test_constructor(self):
            note_grouping = NoteGrouping(notes=[Note.from_dict(working_notes["test1"])], note_model=None, tags=None)

            assert isinstance(note_grouping, NoteGrouping)
            assert isinstance(note_grouping.notes, List)

        def test_from_dict(self, note_grouping_fixtures):
            assert isinstance(note_grouping_fixtures, NoteGrouping)

    class TestDeckPartNote:
        def test_constructor(self):
            dpn = Notes(note_groupings=[NoteGrouping.from_dict(working_note_groupings["nothing_grouped"])])
            assert isinstance(dpn, Notes)

        def test_from_dict(self):
            dpn = Notes.from_dict({NOTE_GROUPINGS: [working_note_groupings["nothing_grouped"]]})
            assert isinstance(dpn, Notes)


class TestDumpToYaml:
    @staticmethod
    def _make_temp_file(tmpdir):
        folder = tmpdir.mkdir("yaml_files")
        file = folder.join("test.yaml")
        file.write("test")
        return file

    class TestNote:
        @staticmethod
        def _assert_dump_to_yaml(tmpdir, ystring, note_name):
            file = TestDumpToYaml._make_temp_file(tmpdir)

            note = Note.from_dict(working_notes[note_name])
            note.dump_to_yaml(str(file))

            assert file.read() == ystring

        def test_all1(self, tmpdir):
            ystring = dedent(f'''\
            {FIELDS}:
            - first
            {GUID}: '12345'
            {NOTE_MODEL}: model_name
            {TAGS}:
            - noun
            - other
            ''')

            self._assert_dump_to_yaml(tmpdir, ystring, "test1")

        def test_all2(self, tmpdir):
            ystring = dedent(f'''\
            {FIELDS}:
            - english
            - german
            {GUID}: sdfhfghsvsdv
            {NOTE_MODEL}: LL Test
            {TAGS}:
            - marked
            ''')

            self._assert_dump_to_yaml(tmpdir, ystring, "test2")

        def test_no_note_model(self, tmpdir):
            ystring = dedent(f'''\
            {FIELDS}:
            - first
            {GUID}: '12345'
            {TAGS}:
            - noun
            - other
            ''')

            self._assert_dump_to_yaml(tmpdir, ystring, "no_note_model")

        def test_no_tags(self, tmpdir):
            for num, note in enumerate(["no_tags1", "no_tags2"]):
                ystring = dedent(f'''\
                {FIELDS}:
                - first
                {GUID}: '12345'
                {NOTE_MODEL}: model_name
                ''')

                self._assert_dump_to_yaml(tmpdir.mkdir(str(num)), ystring, note)

        def test_with_flags(self, tmpdir):
            ystring = dedent(f'''\
            {FIELDS}:
            - first
            {GUID}: '12345'
            {FLAGS}: 1
            {NOTE_MODEL}: model_name
            {TAGS}:
            - noun
            - other
            ''')

            self._assert_dump_to_yaml(tmpdir, ystring, "test1_with_flags")

        def test_with_default_flags(self, tmpdir):
            ystring = dedent(f'''\
            {FIELDS}:
            - first
            {GUID}: '12345'
            {NOTE_MODEL}: model_name
            {TAGS}:
            - noun
            - other
            ''')

            self._assert_dump_to_yaml(tmpdir, ystring, "test1_with_default_flags")

    class TestNoteGrouping:
        @staticmethod
        def _assert_dump_to_yaml(tmpdir, ystring, note_grouping_name):
            file = TestDumpToYaml._make_temp_file(tmpdir)

            note = NoteGrouping.from_dict(working_note_groupings[note_grouping_name])
            note.dump_to_yaml(str(file))

            assert file.read() == ystring

        def test_nothing_grouped(self, tmpdir):
            ystring = dedent(f'''\
            {NOTES}:
            - {FIELDS}:
              - first
              {GUID}: '12345'
              {NOTE_MODEL}: model_name
              {TAGS}:
              - noun
              - other
            - {FIELDS}:
              - english
              - german
              {GUID}: sdfhfghsvsdv
              {NOTE_MODEL}: LL Test
              {TAGS}:
              - marked
            ''')

            self._assert_dump_to_yaml(tmpdir, ystring, "nothing_grouped")

        def test_note_model_grouped(self, tmpdir):
            ystring = dedent(f'''\
            {NOTE_MODEL}: model_name
            {NOTES}:
            - {FIELDS}:
              - first
              {GUID}: '12345'
              {TAGS}:
              - noun
              - other
            - {FIELDS}:
              - second
              {GUID}: '67890'
              {TAGS}:
              - noun
              - other
            ''')

            self._assert_dump_to_yaml(tmpdir, ystring, "note_model_grouped")

        def test_note_tags_grouped(self, tmpdir):
            ystring = dedent(f'''\
            {TAGS}:
            - noun
            - other
            {NOTES}:
            - {FIELDS}:
              - first
              {GUID}: '12345'
              {NOTE_MODEL}: model_name
            - {FIELDS}:
              - first
              {GUID}: '12345'
              {NOTE_MODEL}: model_name
            ''')

            self._assert_dump_to_yaml(tmpdir, ystring, "tags_grouped")

        def test_note_model_and_tags_grouped(self, tmpdir):
            ystring = dedent(f'''\
            {NOTE_MODEL}: model_name
            {TAGS}:
            - noun
            - other
            {NOTES}:
            - {FIELDS}:
              - first
              {GUID}: '12345'
            - {FIELDS}:
              - first
              {GUID}: '12345'
            ''')

            self._assert_dump_to_yaml(tmpdir, ystring, "model_and_tags_grouped")

    class TestDeckPartNotes:
        @staticmethod
        def _assert_dump_to_yaml(tmpdir, ystring, groups: list):
            file = TestDumpToYaml._make_temp_file(tmpdir)

            note = Notes.from_dict({NOTE_GROUPINGS: [working_note_groupings[name] for name in groups]})
            note.dump_to_yaml(str(file))

            assert file.read() == ystring

        def test_two_groupings(self, tmpdir):
            ystring = dedent(f'''\
            {NOTE_GROUPINGS}:
            - {NOTE_MODEL}: model_name
              {TAGS}:
              - noun
              - other
              {NOTES}:
              - {FIELDS}:
                - first
                {GUID}: '12345'
              - {FIELDS}:
                - first
                {GUID}: '12345'
            - {NOTES}:
              - {FIELDS}:
                - first
                {GUID}: '12345'
                {NOTE_MODEL}: model_name
                {TAGS}:
                - noun
                - other
              - {FIELDS}:
                - english
                - german
                {GUID}: sdfhfghsvsdv
                {NOTE_MODEL}: LL Test
                {TAGS}:
                - marked
            ''')

            self._assert_dump_to_yaml(tmpdir, ystring, ["model_and_tags_grouped", "nothing_grouped"])


class TestFunctionality:
    class TestGetMediaReferences:
        class TestNote:
            @pytest.mark.parametrize("fields, expected_count", [
                ([], 0),
                (["nothing", "empty", "can't find nothing here"], 0),
                (["<img src=\"animal.jpg\">", "empty", "can't find nothing here"], 1),
                (["<img src=\"animal.jpg\">", "<img src=\"animal.jpg\">", "<img src=\"animal.jpg\">"], 1),
                (["<img src=\"animal.jpg\">", "<img src=\"food.jpg\">", "<img src=\"object.jpg\">"], 3),
                (["<img src=\"animal.jpg\">", "[sound:test.mp3]", "[sound:test.mp3]"], 2),
            ])
            def test_all(self, fields, expected_count):
                note = Note(fields=fields, note_model=None, guid="", tags=None, flags=0)
                media_found = note.get_all_media_references()
                assert isinstance(media_found, Set)
                assert len(media_found) == expected_count

    class TestGetAllNoteModels:
        class TestNoteGrouping:
            def test_nothing_grouped(self):
                group = NoteGrouping.from_dict(working_note_groupings["nothing_grouped"])
                models = group.get_all_known_note_model_names()
                assert models == {'LL Test', 'model_name'}

            def test_grouped(self):
                group = NoteGrouping.from_dict(working_note_groupings["note_model_grouped"])
                models = group.get_all_known_note_model_names()
                assert models == {'model_name'}

        class TestDeckPartNotes:
            def test_two_groups_two_models(self):
                dpn = Notes.from_dict(working_dpns["two_groups_two_models"])
                models = dpn.get_all_known_note_model_names()
                assert models == {'LL Test', 'model_name'}

            def test_two_groups_three_models(self):
                dpn = Notes.from_dict(working_dpns["two_groups_three_models"])
                models = dpn.get_all_known_note_model_names()
                assert models == {'LL Test', 'model_name', 'different_model'}

    # class TestGetAllNotes:
    #     class TestNoteGrouping:
    #         def test_nothing_grouped(self):
    #             group = NoteGrouping.from_dict(working_note_groupings["nothing_grouped"])
    #             notes = group.get_all_notes_copy([], False)
    #             assert len(notes) == 2
    #
    #         def test_model_grouped(self):
    #             group = NoteGrouping.from_dict(working_note_groupings["note_model_grouped"])
    #             assert group.note_model == "model_name"
    #             assert all([note.note_model is None for note in group.notes])
    #
    #             notes = group.get_all_notes_copy()
    #             assert {note.note_model for note in notes} == {"model_name"}
    #
    #         def test_tags_grouped(self):
    #             group = NoteGrouping.from_dict(working_note_groupings["tags_grouped"])
    #             assert group.tags == ["noun", "other"]
    #             assert all([note.tags is None or note.tags == [] for note in group.notes])
    #
    #             notes = group.get_all_notes_copy()
    #             assert all([note.tags == ["noun", "other"] for note in notes])
    #
    #         def test_tags_grouped_as_addition(self):
    #             group = NoteGrouping.from_dict(working_note_groupings["tags_grouped_as_addition"])
    #             assert group.tags == ["test", "recent"]
    #             assert all([note.tags is not None for note in group.notes])
    #
    #             notes = group.get_all_notes_copy()
    #             assert notes[0].tags == ['noun', 'other', "test", "recent"]
    #             assert notes[1].tags == ['marked', "test", "recent"]
    #
    #         def test_no_tags(self):
    #             group = NoteGrouping.from_dict(working_note_groupings["tags_grouped"])
    #             group.tags = None
    #             assert all([note.tags is None or note.tags == [] for note in group.notes])
    #
    #             notes = group.get_all_notes_copy()
    #             assert all([note.tags == [] for note in notes])
    #



import re
from abc import ABCMeta, abstractmethod
from dataclasses import dataclass
from textwrap import indent
from typing import Dict, List, Type, Set

from brain_brew.commands.run_recipe.build_task import BuildTask
from brain_brew.representation.yaml.yaml_object import YamlObject


@dataclass
class RecipeBuilder(YamlObject, metaclass=ABCMeta):
    tasks: List[BuildTask]

    def tasks_to_encoded(self) -> list:
        return [{task.task_name(): task.encode_rep()} for task in self.tasks]

    @classmethod
    def from_list(cls, data: List[dict]):
        tasks = cls.read_tasks(data)
        return cls(
            tasks=tasks
        )

    @classmethod
    @abstractmethod
    def known_task_dict(cls) -> Dict[str, Type[BuildTask]]:
        pass

    @classmethod
    def build_yamale_root_node(cls, subclasses: Set[Type['BuildTask']]) -> str:
        task_list = []
        for c in sorted(subclasses, key=lambda x: x.task_name()):
            task_command = f"any(include('{c.task_name()}'), list(include('{c.task_name()}')))"\
                if c.accepts_list_of_self else f"include('{c.task_name()}')"
            task_list.append(f"map({task_command}, key=regex('{c.task_regex()}', ignore_case=True))")

        final_tasks: str = "list(\n" + indent(",\n".join(task_list), '  ') + "\n)\n"

        return final_tasks

    @classmethod
    def read_tasks(cls, tasks: List[dict]) -> list:
        task_regex_matches = cls.known_task_dict()
        build_tasks = []

        def find_matching_task(task_n):
            for regex, task_to_run in task_regex_matches.items():
                if re.match(regex, task_n, re.RegexFlag.IGNORECASE):
                    return task_to_run
            return None

        # Tasks
        for task in tasks:
            task_keys = list(task.keys())
            if len(task_keys) != 1:
                raise KeyError(f"Task should only contain 1 entry, but contains {task_keys} instead. "
                               f"Missing list separator '-'?", task)

            task_name = task_keys[0]
            task_arguments = task[task_keys[0]]

            matching_task = find_matching_task(task_name)
            if matching_task is not None:
                if matching_task.accepts_list_of_self and isinstance(task_arguments, list):
                    task_or_tasks = [matching_task.from_repr(t_arg) for t_arg in task_arguments]
                else:
                    task_or_tasks = [matching_task.from_repr(task_arguments)]

                for inner_task in task_or_tasks:
                    build_tasks.append(inner_task)
                    if inner_task.execute_immediately:
                        inner_task.execute()
            else:
                raise KeyError(f"Unknown task '{task_name}'")  # TODO: check this first on all and return all errors

        return build_tasks

    def execute(self):
        for task in self.tasks:
            if not task.execute_immediately:
                task.execute()



import inspect
import logging


class RepresentationBase:
    @classmethod
    def from_dict(cls, data: dict):
        expected_values = {
            k: v for k, v in data.items()
            if k in inspect.signature(cls).parameters
        }

        if len(expected_values) != len(data):
            logging.warning(f"Unexpected values found when creating '{cls.__name__}': "
                            f"{[k for k, v in data.items() if k not in list(expected_values.keys())]}"
                            "\n!!! Please report this error if it seems strange")

        return cls(**expected_values)

    def encode(self):
        return {key: value for key, value in self.__dict__.items() if self.encode_filter(key, value)}

    def encode_filter(self, key, value):
        if value is None:
            return False
        if not value:
            return False
        return True



import setuptools
from brain_brew.front_matter import latest_version_number

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="Brain-Brew",
    version=latest_version_number(),
    author="Jordan Munch O'Hare",
    author_email="brainbrew@jordan.munchohare.com",
    description="Automated Anki flashcard creation and extraction to/from Csv ",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/ohare93/brain-brew",
    packages=setuptools.find_packages(),
    include_package_data=True,
    entry_points={
        'console_scripts': [
            'brain_brew = brain_brew.main:main',
            'brain-brew = brain_brew.main:main',
            'brainbrew = brain_brew.main:main',
        ]
    },
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: Public Domain",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.7',
    install_requires=[
        'ruamel.yaml.clib>=0.2.2',
        'ruamel.yaml>=0.16.10',
        'yamale>=3.0.4'
    ]
)



import os
from dataclasses import dataclass, field
from typing import Optional, Union, Set

from brain_brew.configuration.anki_field import AnkiField
from brain_brew.configuration.representation_base import RepresentationBase
from brain_brew.interfaces.yamale_verifyable import YamlRepr
from brain_brew.representation.generic.html_file import HTMLFile
from brain_brew.representation.yaml.yaml_object import YamlObject
from brain_brew.utils import find_media_in_field, split_by_regex

NAME = AnkiField("name")
ORDINAL = AnkiField("ord", "ordinal")
QUESTION_FORMAT = AnkiField("qfmt", "question_format")
ANSWER_FORMAT = AnkiField("afmt", "answer_format")
BROWSER_ANSWER_FORMAT = AnkiField("bafmt", "browser_answer_format", default_value="")
BROWSER_QUESTION_FORMAT = AnkiField("bqfmt", "browser_question_format", default_value="")
DECK_OVERRIDE_ID = AnkiField("did", "deck_override_id", default_value=None)
BROWSER_FONT = AnkiField("bfont", "browser_font", default_value="")
BROWSER_FONT_SIZE = AnkiField("bsize", "browser_font_size", default_value=0)
SCRATCH_PAD = AnkiField("scratchPad", "scratch_pad", default_value=0)

HTML_FILE = AnkiField("html_file")
BROWSER_HTML_FILE = AnkiField("browser_html_file", default_value=None)

html_separator_regex = r'[\n]{1,}[-]{1,}[\n]{1,}'


@dataclass
class Template(RepresentationBase, YamlObject, YamlRepr):
    @classmethod
    def task_name(cls) -> str:
        return r'note_model_template_from_html'

    @classmethod
    def yamale_schema(cls) -> str:
        return f"""\
            name: str()
            html_file: str()
            browser_html_file: str(required=False)
            deck_override_id: int(required=False)
        """

    @dataclass
    class HTML(RepresentationBase):
        name: str
        html_file: str
        browser_html_file: Optional[str] = field(default=None)
        browser_font: str = field(default=BROWSER_FONT.default_value)
        browser_font_size: int = field(default=BROWSER_FONT_SIZE.default_value)
        deck_override_id: Optional[int] = field(default=DECK_OVERRIDE_ID.default_value)
        scratch_pad: int = field(default=SCRATCH_PAD.default_value)

    @classmethod
    def from_repr(cls, data: Union[HTML, dict]):
        rep: cls.HTML = data if isinstance(data, cls.HTML) else cls.HTML.from_dict(data)
        return cls.from_html_files(rep)

    @classmethod
    def from_yaml_file(cls, filename: str) -> 'Template':
        return cls.from_dict(cls.read_to_dict(filename))

    @dataclass
    class CrowdAnki(RepresentationBase):
        name: str
        qfmt: str
        afmt: str
        bqfmt: str = field(default=BROWSER_QUESTION_FORMAT.default_value)
        bafmt: str = field(default=BROWSER_ANSWER_FORMAT.default_value)
        bfont: str = field(default=BROWSER_FONT.default_value)
        bsize: int = field(default=BROWSER_FONT_SIZE.default_value)
        ord: int = field(default=None)
        did: Optional[int] = field(default=None)
        scratchPad: int = field(default=SCRATCH_PAD.default_value)

    name: str
    question_format: str
    answer_format: str
    question_format_in_browser: str = field(default=BROWSER_QUESTION_FORMAT.default_value)
    answer_format_in_browser: str = field(default=BROWSER_ANSWER_FORMAT.default_value)
    browser_font: str = field(default=BROWSER_FONT.default_value)
    browser_font_size: int = field(default=BROWSER_FONT_SIZE.default_value)
    deck_override_id: Optional[int] = field(default=DECK_OVERRIDE_ID.default_value)
    scratch_pad: int = field(default=SCRATCH_PAD.default_value)

    html_file: Optional[str] = field(default="")
    browser_html_file: Optional[str] = field(default="")

    @classmethod
    def from_html_files(cls, data: Union[HTML, dict]):
        html_rep: cls.HTML = data if isinstance(data, cls.HTML) else cls.HTML.from_dict(data)

        html_file = HTMLFile.create_or_get(html_rep.html_file)
        browser_html_file = HTMLFile.create_or_get(html_rep.browser_html_file) if html_rep.browser_html_file else None

        main_data = html_file.get_data(deep_copy=True)
        browser_data = browser_html_file.get_data(deep_copy=True) if browser_html_file else None

        def split_template(the_data, file):
            split = split_by_regex(the_data, html_separator_regex)
            if len(split) != 2:
                raise ValueError(f"Cannot find" if len(split) < 2 else "More than one"
                                                                       f" separator '---' in html file '{file.file_location}'")
            return split[0], split[1]

        front, back = split_template(main_data, html_file)
        browser_front, browser_back = split_template(browser_data, browser_html_file) if browser_data else ("", "")

        return cls(
            name=html_rep.name,
            question_format=front,
            answer_format=back,
            question_format_in_browser=browser_front,
            answer_format_in_browser=browser_back,
            deck_override_id=html_rep.deck_override_id,
            html_file=html_rep.html_file,
            browser_html_file=html_rep.browser_html_file,
            browser_font=html_rep.browser_font,
            browser_font_size=html_rep.browser_font_size,
            scratch_pad=html_rep.scratch_pad,
        )

    @classmethod
    def from_crowdanki(cls, data: Union[CrowdAnki, dict]):
        ca: cls.CrowdAnki = data if isinstance(data, cls.CrowdAnki) else cls.CrowdAnki.from_dict(data)
        return cls(
            name=ca.name, question_format=ca.qfmt, answer_format=ca.afmt,
            question_format_in_browser=ca.bqfmt, answer_format_in_browser=ca.bafmt,
            deck_override_id=ca.did, browser_font=ca.bfont, browser_font_size=ca.bsize, scratch_pad=ca.scratchPad,
        )

    def encode_as_part(self):
        data_dict = {
            NAME.name: self.name,
            HTML_FILE.name: ""
        }

        if self.has_browser_template():
            data_dict.setdefault(BROWSER_HTML_FILE.name, "")

        DECK_OVERRIDE_ID.append_name_if_differs(data_dict, self.deck_override_id)
        BROWSER_FONT.append_name_if_differs(data_dict, self.browser_font)
        BROWSER_FONT_SIZE.append_name_if_differs(data_dict, self.browser_font_size)
        SCRATCH_PAD.append_name_if_differs(data_dict, self.scratch_pad)

        return data_dict

    def encode_as_crowdanki(self, ordinal: int) -> dict:
        data_dict = {
            ANSWER_FORMAT.anki_name: self.answer_format,
            BROWSER_ANSWER_FORMAT.anki_name: self.answer_format_in_browser,
            BROWSER_FONT.anki_name: self.browser_font,
            BROWSER_QUESTION_FORMAT.anki_name: self.question_format_in_browser,
            BROWSER_FONT_SIZE.anki_name: self.browser_font_size,
            DECK_OVERRIDE_ID.anki_name: self.deck_override_id,
            NAME.anki_name: self.name,
            ORDINAL.anki_name: ordinal,
            QUESTION_FORMAT.anki_name: self.question_format,
            SCRATCH_PAD.anki_name: self.scratch_pad,
        }

        return data_dict

    def encode(self) -> dict:
        data_dict = {
            NAME.name: self.name,
            QUESTION_FORMAT.name: self.question_format,
            ANSWER_FORMAT.name: self.answer_format
        }

        BROWSER_QUESTION_FORMAT.append_name_if_differs(data_dict, self.question_format_in_browser)
        BROWSER_ANSWER_FORMAT.append_name_if_differs(data_dict, self.answer_format_in_browser)
        DECK_OVERRIDE_ID.append_name_if_differs(data_dict, self.deck_override_id)
        BROWSER_FONT.append_name_if_differs(data_dict, self.browser_font)
        BROWSER_FONT_SIZE.append_name_if_differs(data_dict, self.browser_font_size)
        SCRATCH_PAD.append_name_if_differs(data_dict, self.scratch_pad)

        return data_dict

    def get_all_media_references(self) -> Set[str]:
        all_media = set() \
            .union(find_media_in_field(self.question_format)) \
            .union(find_media_in_field(self.answer_format)) \
            .union(find_media_in_field(self.question_format_in_browser)) \
            .union(find_media_in_field(self.answer_format_in_browser))
        return all_media

    def has_browser_template(self):
        return BROWSER_QUESTION_FORMAT.does_differ(self.question_format_in_browser) \
                or BROWSER_ANSWER_FORMAT.does_differ(self.answer_format_in_browser)

    def get_template_files_data(self):
        template = f"{self.question_format}\n\n--\n\n{self.answer_format}"
        browser_template = f"{self.question_format}\n\n--\n\n{self.answer_format}" if self.has_browser_template() else None

        return template, browser_template



from dataclasses import dataclass, field
from typing import Optional, Union, List

from brain_brew.build_tasks.crowd_anki.shared_base_notes import SharedBaseNotes
from brain_brew.build_tasks.overrides.notes_override import NotesOverride
from brain_brew.configuration.part_holder import PartHolder
from brain_brew.configuration.representation_base import RepresentationBase
from brain_brew.interfaces.yamale_verifyable import YamlRepr
from brain_brew.representation.json.wrappers_for_crowd_anki import CrowdAnkiNoteWrapper
from brain_brew.representation.yaml.notes import Notes, Note
from brain_brew.utils import blank_str_if_none


@dataclass
class NotesToCrowdAnki(YamlRepr, SharedBaseNotes):
    @classmethod
    def task_name(cls) -> str:
        return r'notes_to_crowd_anki'

    @classmethod
    def yamale_schema(cls) -> str:
        return f'''\
            part_id: str()
            sort_order: list(str(), required=False)
            reverse_sort: bool(required=False)
            additional_items_to_add: map(str(), key=str(), required=False)
            override: include('{NotesOverride.task_name()}', required=False)
            case_insensitive_sort: bool(required=False)
        '''

    @classmethod
    def yamale_dependencies(cls) -> set:
        return {NotesOverride}

    @dataclass
    class Representation(RepresentationBase):
        part_id: str
        additional_items_to_add: Optional[dict] = field(default_factory=lambda: None)
        sort_order: Optional[List[str]] = field(default_factory=lambda: None)
        reverse_sort: Optional[bool] = field(default_factory=lambda: None)
        override: Optional[dict] = field(default_factory=lambda: None)
        case_insensitive_sort: Optional[bool] = field(default_factory=lambda: None)

    @classmethod
    def from_repr(cls, data: Union[Representation, dict]):
        rep: cls.Representation = data if isinstance(data, cls.Representation) else cls.Representation.from_dict(data)
        return cls(
            rep=rep,
            notes=PartHolder.from_file_manager(rep.part_id).part,
            sort_order=SharedBaseNotes._get_sort_order(rep.sort_order),
            reverse_sort=SharedBaseNotes._get_reverse_sort(rep.reverse_sort),
            additional_items_to_add=rep.additional_items_to_add or {},
            override=NotesOverride.from_repr(rep.override) if rep.override else None,
            case_insensitive_sort=rep.case_insensitive_sort or True
        )

    rep: Representation
    notes: Notes
    additional_items_to_add: dict
    sort_order: Optional[List[str]] = field(default_factory=lambda: None)
    reverse_sort: Optional[bool] = field(default_factory=lambda: None)
    override: Optional[NotesOverride] = field(default_factory=lambda: None)
    case_insensitive_sort: bool = field(default=True)

    def execute(self, nm_name_to_id: dict) -> List[dict]:

        notes = self.notes.get_sorted_notes_copy(
            sort_by_keys=self.sort_order,
            reverse_sort=self.reverse_sort,
            case_insensitive_sort=self.case_insensitive_sort
        )

        if self.override:
            notes = [self.override.override(note) for note in notes]

        note_dicts = [self.note_to_ca_note(note, nm_name_to_id, self.additional_items_to_add) for note in notes]

        return note_dicts

    @staticmethod
    def note_to_ca_note(note: Note, nm_name_to_id: dict, additional_items_to_add: dict) -> dict:
        wrapper = CrowdAnkiNoteWrapper({
            "__type__": "Note",
            "data": ""
        })

        for key, value in additional_items_to_add.items():
            wrapper.data[key] = blank_str_if_none(value)

        wrapper.fields = note.fields
        wrapper.flags = note.flags
        wrapper.guid = note.guid
        wrapper.note_model = nm_name_to_id[note.note_model]
        wrapper.tags = note.tags

        return wrapper.data



from dataclasses import dataclass
from typing import Union

from brain_brew.build_tasks.deck_parts.media_group_from_folder import MediaGroupFromFolder
from brain_brew.configuration.part_holder import PartHolder
from brain_brew.representation.json.crowd_anki_export import CrowdAnkiExport
from brain_brew.representation.yaml.media_group import MediaGroup
from brain_brew.transformers.create_media_group_from_location import create_media_group_from_location


@dataclass
class MediaGroupFromCrowdAnki(MediaGroupFromFolder):
    @classmethod
    def task_name(cls) -> str:
        return r"media_group_from_crowd_anki"

    @classmethod
    def from_repr(cls, data: Union[MediaGroupFromFolder.Representation, dict]):
        rep: cls.Representation = data if isinstance(data, cls.Representation) else cls.Representation.from_dict(data)

        cae: CrowdAnkiExport = CrowdAnkiExport.create_or_get(rep.source)
        return cls(
            rep=rep,
            part=create_media_group_from_location(
                part_id=rep.part_id,
                save_to_file=rep.save_to_file,
                media_group=MediaGroup.from_directory(cae.media_loc, rep.recursive),
                groups_to_blacklist=list(holder.part for holder in
                                         map(PartHolder.from_file_manager, rep.filter_blacklist_from_parts)),
                groups_to_whitelist=list(holder.part for holder in
                                         map(PartHolder.from_file_manager, rep.filter_whitelist_from_parts))
            )
        )

    rep: MediaGroupFromFolder.Representation
    part: MediaGroup

    def execute(self):
        pass
